%% Grfalgoritmusokat tartalmaz modul.
%% A grfalgoritmusok sokat hasznljk a common modul segdfggvnyeit.

-module(algorithm).

-export([depthWalk/2, breadthWalk/2]).
-export([dijkstra/2, prim/2, bellman_ford/2]).

%% Rekordok.
%% A vertex_label egy cscs cmkje, amely lerja az llapott egy algoritmus aktulis fzisban. Az algoritmusok s a segdfggvnyek hasznljk.
-include("record_vertex_label.hrl").
%% A vertex_info egy cscsrl egy algoritmus elvgzse utn informcit nyjt rekord. Tbb algoritmus kimenete ilyen rekordok listja.
-include("record_vertex_info.hrl").

%% Mlysgi bejrs algoritmusa.
%% Az algoritmus bemeneti paramterl egy grfot s annak egy Start cscst vrja.
%% Az algoritmus visszatrsi rtke a cscsok listja bejrsi sorrendben.
depthWalk(Graph, Node) ->
    depthWalk(Graph, [], [Node]).

depthWalk(_, _, []) ->
    [];

% Az algoritmus bels megvalstsban 3 paramteres:
% - a grf,
% - a mr feldolgozott cscsok,
% - a mg feldolgozsra vr cscsok.
depthWalk(Graph, Done, [Node | Wait]) ->
    % Az algoritmus egy lpsben veszi az els feldolgozsra vr cscsot s ha mg nem lett feldolgzozva, akkor tteszi a mr feldolgozottak kz.
    % Tovbb a belle elrhet cscsokat beteszi a feldolgozsra vr cscsok listjnak elejre.
    Contains = common:contains(Done, Node),
    if
	not Contains ->
	    [Node | depthWalk(Graph, [Node | Done], lists:append(digraph:out_neighbours(Graph, Node), Wait))];
	true ->
	    depthWalk(Graph, Done, Wait)
    end.

%% Szlessgi bejrs algoritmusa.
%% Az algoritmus bemeneti paramterl egy grfot s annak egy Start cscst vrja.
%% Az algoritmus visszatrsi rtke a cscsok listja bejrsi sorrendben.
breadthWalk(Graph, Node) ->
    breadthWalk(Graph, [], [Node]).

breadthWalk(_, _, []) ->
    [];

% Az algoritmus bels megvalstsban 3 paramteres:
% - a grf,
% - a mr feldolgozott cscsok,
% - a mg feldolgozsra vr cscsok.
breadthWalk(Graph, Done, [Node | Wait]) ->
    % Az algoritmus egy lpsben veszi az els feldolgozsra vr cscsot s ha mg nem lett feldolgzozva, akkor tteszi a mr feldolgozottak kz.
    % Tovbb a belle elrhet cscsokat beteszi a feldolgozsra vr cscsok listjnak elejre vgre.
    Contains = common:contains(Done, Node),
    if
	not Contains ->
	    [Node | breadthWalk(Graph, [Node | Done], lists:append(Wait, digraph:out_neighbours(Graph, Node)))];
	true ->
	    breadthWalk(Graph, Done, Wait)
    end.

%% Dijkstra algoritmus.
%% Az algoritmus egy irnytott, nemnegatvan lslyozott grfban egy Start cscsbl kiindulva megadja a tbbi cscshoz vezet minimlis kltsg utat.
%% Bemeneti paramterknt a grfot s a Start cscsot kell megadni.
%% A fggvny a cscsokat ler vertex_info rekordok listjval tr vissza, egy rekord tartalma:
%% - node: a cscs azonostja.
%% - depth: a cscshoz vezet t minimlis kltsge az algoritmus vgeztvel.
%% - parent: a szl cscs, ahonnan a minimlis kltsg t vezet a jelenlegi cscsba.
%% Amennyiben a grf negatv let is tartalmaz, a visszatrsi rtk: false.
dijkstra(Graph, Node) ->
    % Cscs- s lcmkk inicializlsa.
    InitedGraph = common:init_edge_labels(common:init_vertex_labels(Graph, Node), 1),
    % Negatv l keresse.
    HasNegativeEdge = common:has_negative_edge(InitedGraph),
    if
	not HasNegativeEdge ->
	    common:get_vertex_labels(dijkstra(InitedGraph));
	true ->
	    false
    end.

% Egy lpsben kivlasztjuk az aktulisan legkisebb elrsi kltsggel rendelkez, mg nem ksz cscsot.
% Ezt a cscsot kssz tesszk s frisstjk a belle kzvetlen elrhet cscsok kltsgeit, ha azok mg nem kszek s alacsonyabb kltsggel is elrhetek.
dijkstra(Graph) ->
    #vertex_info{node = Node, depth = Depth, parent = Parent} = common:search_min_vertex(Graph),
    if
	% A Node akkor null, ha mr nincsen olyan nem ksz cscs, amelynek elrsi kltsge nem infinite.
	Node /= null ->
	    digraph:add_vertex(Graph, Node, #vertex_label{depth = Depth, parent = Parent, done = true}),
	    Edges = digraph:out_edges(Graph, Node),
	    dijkstra_inner(Graph, Edges),
	    dijkstra(Graph);
	true ->
	    Graph
    end.

dijkstra_inner(Graph, []) ->
    Graph;

dijkstra_inner(Graph, [Edge | Edges]) ->
    {_, From, To, Value} = digraph:edge(Graph, Edge),
    {FNode, #vertex_label{depth = FDepth}} = digraph:vertex(Graph, From),
    {TNode, #vertex_label{depth = TDepth, done = TDone}} = digraph:vertex(Graph, To),
    if
	TDone == false andalso (TDepth > FDepth + Value orelse TDepth == infinite) ->
	    % Az j kltsg a szl cscs kltsge plusz az l cmkje lesz.
	    digraph:add_vertex(Graph, TNode, #vertex_label{depth = FDepth + Value, parent = FNode, done = TDone});
	true  ->
	    ok
    end,
    dijkstra_inner(Graph, Edges).

%% Prim algoritmus.
%% Az algoritmus egy irnytatlan, lslyozott grfban egy Start cscsbl kiindulva meghatrozza a minimlis kltsg fesztft.
%% Bemeneti paramterknt a grfot s a Start cscsot kell megadni.
%% A fggvny a cscsokat ler vertex_info rekordok listjval tr vissza, egy rekord tartalma:
%% - node: a cscs azonostja.
%% - depth: a cscs becsatolsi kltsge az algoritmussal a fesztfba.
%% - parent: a szl cscs, az ahonnan kiindul llel ezt a cscsot becsatoltuk a fesztfba.
%% Amennyiben a grf irnytott let is tartalmaz, a visszatrsi rtk: false.
prim(Graph, Node) ->
    % Cscs- s lcmkk inicializlsa.
    InitedGraph = common:init_edge_labels(common:init_vertex_labels(Graph, Node), 1),
    % Irnytott l keresse.
    HasDirectedEdge = common:has_directed_edge(InitedGraph),
    if
	not HasDirectedEdge ->
	    common:get_vertex_labels(prim(InitedGraph));
	true ->
	    false
    end.

% Egy lpsben kivlasztjuk az aktulisan legkisebb elrsi kltsggel rendelkez, mg nem ksz cscsot.
% Ezt a cscsot kssz tesszk s frisstjk a belle kzvetlen elrhet cscsok kltsgeit, ha azok mg nem kszek s alacsonyabb kltsggel is elrhetek.
prim(Graph) ->
    #vertex_info{node = Node, depth = Depth, parent = Parent} = common:search_min_vertex(Graph),
    if
	% A Node akkor null, ha mr nincsen olyan nem ksz cscs, amelynek elrsi kltsge nem infinite.
	Node /= null ->
	    digraph:add_vertex(Graph, Node, #vertex_label{depth = Depth, parent = Parent, done = true}),
	    Edges = digraph:out_edges(Graph, Node),
	    prim_inner(Graph, Edges),
	    prim(Graph);
	true ->
	    Graph
    end.

prim_inner(Graph, []) ->
    Graph;

prim_inner(Graph, [Edge | Edges]) ->
    {_Edge, From, To, Value} = digraph:edge(Graph, Edge),
    {FNode, _FLabel} = digraph:vertex(Graph, From),
    {TNode, #vertex_label{depth = TDepth, done = TDone}} = digraph:vertex(Graph, To),
    if
	TDone == false andalso (TDepth > Value orelse TDepth == infinite) ->
	    % Az j kltsg az l cmkje lesz.
	    digraph:add_vertex(Graph, TNode, #vertex_label{depth = Value, parent = FNode, done = TDone});
	true  ->
	    ok
    end,
    prim_inner(Graph, Edges).

%% Bellman-Ford algoritmus.
%% Az algoritmus egy irnytott, negatv krt nem tartalmaz lslyozott grfban egy Start cscsbl kiindulva megadja a tbbi cscshoz vezet minimlis kltsg utat.
%% Bemeneti paramterknt a grfot s a Start cscsot kell megadni.
%% A fggvny a cscsokat ler vertex_info rekordok listjval tr vissza, egy rekord tartalma:
%% - node: a cscs azonostja.
%% - depth: a cscshoz vezet t minimlis kltsge az algoritmus vgeztvel.
%% - parent: a szl cscs, ahonnan a minimlis kltsg t vezet a jelenlegi cscsba.
%% Amennyiben a grf negatv krt is tartalmaz, azt az algoritmus szleli s ekkor a visszatrsi rtk: false.
bellman_ford(Graph, Node) ->
    % Cscs- s lcmkk inicializlsa, majd az algoritmus vgrehajtsa.
    BellmanedGraph = bellman_ford_inner(common:init_edge_labels(common:init_vertex_labels(Graph, Node), 0), 1),
    % Ammenyiben az algoritmus egy grffal tr vissza, akkor sikeresen vgre lett hajtva, egybknt logikai a visszatrsi rtk tpusa.
    IsBoolean = is_boolean(BellmanedGraph),
    if
	IsBoolean ->
	    BellmanedGraph;
	true ->
	    common:get_vertex_labels(BellmanedGraph)
    end.

% Az algoritmus egy menetben minden let feldolgoz s frissti minden cscs elrsi kltsgt, amely alacsonyabb kltsggel is elrhet.
% Az algoritmus akkor rt vget, ha egy menetben mr nem volt javts.
% Az algoritmusnak legfeljebb n-1 menetre van szksge, ahol n a cscsok szma. Ha az (ellenrz) n. menetben is trtnt javts, akkor a grf negatv krt tartalmaz. 
bellman_ford_inner(Graph, Round) ->
    VertexCount = digraph:no_vertices(Graph),
    if
	VertexCount >= Round ->
	    Edges = digraph:edges(Graph),
	    Corrections = lists:map(
	      fun(Edge) ->
		      {_Edge, From, To, Value} = digraph:edge(Graph, Edge),
		      {FNode, #vertex_label{depth = FDepth}} = digraph:vertex(Graph, From),
		      {TNode, #vertex_label{depth = TDepth}} = digraph:vertex(Graph, To),
		      if
			  FDepth /= infinite andalso (TDepth > FDepth + Value orelse TDepth == infinite) ->
			      % Az j kltsg a szl cscs kltsge plusz az l cmkje lesz.
			      digraph:add_vertex(Graph, TNode, #vertex_label{depth = FDepth + Value, parent = FNode}),
			      true;
			  true ->
			      false
		      end
	      end,
	      Edges),
	    % Trtnt-e javts?
	    HasCorrection = lists:any(
				 fun(Correction) ->
					 Correction
				 end,
				 Corrections),
	    if
		HasCorrection ->
		    bellman_ford_inner(Graph, Round + 1);
		true ->
		    Graph
	    end;
	true ->
	    false
    end.
